/***
 * Fractal ADL Parser
 * Copyright (C) 2002-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 *
 * Contributor: Matthieu Leclercq (externalize interface loading)
 */

package org.objectweb.fractal.adl.bindings;

import java.util.Map;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.interfaces.IDLException;
import org.objectweb.fractal.adl.interfaces.IDLLoader;
import org.objectweb.fractal.adl.interfaces.Interface;
import org.objectweb.fractal.adl.types.TypeInterface;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;

/**
 * An extended {@link BindingLoader} for components with typed interfaces. In
 * addition to the checks performed by {@link BindingLoader}, this
 * implementation checks that from interfaces are client interfaces, that to
 * interfaces are server interfaces, and that the signatures and contingency of
 * client and server interfaces are compatible.
 */

public class TypeBindingLoader extends BindingLoader {

  // --------------------------------------------------------------------------
  // Client interfaces
  // --------------------------------------------------------------------------

  /** The name of the {@link #interfaceCodeLoaderItf} client interface. */
  public static final String INTERFACE_LOADER_BINDING = "interface-loader";

  /** The optional client interface used to load interface signatures. */
  public IDLLoader           interfaceCodeLoaderItf;

  // --------------------------------------------------------------------------
  // Overridden methods
  // --------------------------------------------------------------------------

  @Override
  protected void checkBinding(final Binding binding, final Interface fromItf,
      final Interface toItf, final Map<Object, Object> context)
      throws ADLException {
    if (fromItf instanceof TypeInterface && toItf instanceof TypeInterface) {
      final TypeInterface cItf = (TypeInterface) fromItf;
      final TypeInterface sItf = (TypeInterface) toItf;
      if (binding.getFrom().startsWith("this.")) {
        if (!TypeInterface.SERVER_ROLE.equals(cItf.getRole())) {
          throw new ADLException(
              "Invalid binding: 'from' interface is not an internal client interface",
              binding);
        }
      } else {
        if (!TypeInterface.CLIENT_ROLE.equals(cItf.getRole())) {
          throw new ADLException(
              "Invalid binding: 'from' interface is not a client interface",
              binding);
        }
      }
      if (binding.getTo().startsWith("this.")) {
        if (!TypeInterface.CLIENT_ROLE.equals(sItf.getRole())) {
          throw new ADLException(
              "Invalid binding: 'to' interface is not an internal server interface",
              binding);
        }
      } else {
        if (!TypeInterface.SERVER_ROLE.equals(sItf.getRole())) {
          throw new ADLException(
              "Invalid binding: 'to' interface is not a server interface",
              binding);
        }
      }

      boolean cIsMandatory = true;
      if (TypeInterface.OPTIONAL_CONTINGENCY.equals(cItf.getContingency())) {
        cIsMandatory = false;
      }
      boolean sIsMandatory = true;
      if (TypeInterface.OPTIONAL_CONTINGENCY.equals(sItf.getContingency())) {
        sIsMandatory = false;
      }
      if (cIsMandatory && !sIsMandatory) {
        throw new ADLException(
            "Invalid binding: binding from mandatory to optional interface",
            binding);
      }

      if (interfaceCodeLoaderItf != null) {
        try {
          final Object cClass = interfaceCodeLoaderItf.loadInterface(cItf
              .getSignature(), context);
          final Object sClass = interfaceCodeLoaderItf.loadInterface(sItf
              .getSignature(), context);
          if ((cClass instanceof Class) && (sClass instanceof Class)
              && !((Class<?>) cClass).isAssignableFrom((Class<?>) sClass)) {
            throw new ADLException("Invalid binding: incompatible signatures",
                binding);
          }
        } catch (final IDLException e) {
        }
      }
    }
  }

  @Override
  protected Interface getInterface(final String name,
      final Map<String, Interface> itfs) {
    Interface itf = super.getInterface(name, itfs);
    if (itf == null) {
      for (final Map.Entry<String, Interface> entry : itfs.entrySet()) {
        final String n = entry.getKey();
        itf = itfs.get(n);
        if (itf instanceof TypeInterface) {
          final TypeInterface typeItf = (TypeInterface) itf;
          final String cardinality = typeItf.getCardinality();
          if (TypeInterface.COLLECTION_CARDINALITY.equals(cardinality)) {
            if (name.startsWith(n)) {
              return itf;
            }
          }
        }
      }
      return null;
    } else {
      return itf;
    }
  }

  // ------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ------------------------------------------------------------------------

  @Override
  public void bindFc(final String s, final Object o)
      throws NoSuchInterfaceException, IllegalBindingException {

    if (s == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (INTERFACE_LOADER_BINDING.equals(s)) {
      interfaceCodeLoaderItf = (IDLLoader) o;
    } else {
      super.bindFc(s, o);
    }
  }

  @Override
  public String[] listFc() {
    final String[] superList = super.listFc();
    final String[] list = new String[superList.length + 1];
    list[0] = INTERFACE_LOADER_BINDING;
    System.arraycopy(superList, 0, list, 1, superList.length);
    return list;
  }

  @Override
  public Object lookupFc(final String s) throws NoSuchInterfaceException {
    if (s == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (INTERFACE_LOADER_BINDING.equals(s)) {
      return interfaceCodeLoaderItf;
    } else {
      return super.lookupFc(s);
    }
  }

  @Override
  public void unbindFc(final String s) throws IllegalBindingException,
      NoSuchInterfaceException {

    if (s == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (INTERFACE_LOADER_BINDING.equals(s)) {
      interfaceCodeLoaderItf = null;
    } else {
      super.unbindFc(s);
    }
  }

}
