/***
 * Fractal ADL Parser
 * Copyright (C) 2002-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 *
 * Contributor(s): Philippe Merle
 *                 Matthieu Leclercq (externalize interface and implementation
 *   loading)
 */

package org.objectweb.fractal.adl.implementations;

import java.util.Map;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.AbstractLoader;
import org.objectweb.fractal.adl.Definition;
import org.objectweb.fractal.adl.Loader;
import org.objectweb.fractal.adl.attributes.Attributes;
import org.objectweb.fractal.adl.attributes.AttributesContainer;
import org.objectweb.fractal.adl.components.Component;
import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.adl.interfaces.IDLLoader;
import org.objectweb.fractal.adl.interfaces.Interface;
import org.objectweb.fractal.adl.interfaces.InterfaceContainer;
import org.objectweb.fractal.adl.types.TypeInterface;
import org.objectweb.fractal.api.NoSuchInterfaceException;

/**
 * A {@link org.objectweb.fractal.adl.Loader} to check {@link Implementation}
 * nodes in definitions. This loader checks that the Java classes specified in
 * these nodes exist.
 */

public class ImplementationLoader extends AbstractLoader {

  // --------------------------------------------------------------------------
  // Client interfaces
  // --------------------------------------------------------------------------

  public static final String      INTERFACE_LOADER_BINDING      = "interface-loader";

  public static final String      IMPLEMENTATION_LOADER_BINDING = "implementation-code-loader";

  /** The {@link IDLLoader} used by this loader to load interface code. */

  public IDLLoader                interfaceLoaderItf;

  /**
   * The {@link ImplementationCodeLoader} used by this loader to load
   * implementation code.
   */

  public ImplementationCodeLoader implementationCodeLoaderItf;

  // --------------------------------------------------------------------------
  // Implementation of the Loader interface
  // --------------------------------------------------------------------------

  public Definition load(final String name, final Map<Object, Object> context)
      throws ADLException {
    final Definition d = clientLoader.load(name, context);
    checkNode(d, context);
    return d;
  }

  // --------------------------------------------------------------------------
  // Checking methods
  // --------------------------------------------------------------------------

  protected void checkNode(final Object node, final Map<Object, Object> context)
      throws ADLException {
    if (node instanceof ImplementationContainer) {
      checkImplementationContainer((ImplementationContainer) node, context);
    }
    if (node instanceof ControllerContainer) {
      checkControllerContainer((ControllerContainer) node);
    }
    if (node instanceof ComponentContainer) {
      for (final Component comp : ((ComponentContainer) node).getComponents()) {
        checkNode(comp, context);
      }
    }
  }

  protected void checkImplementationContainer(
      final ImplementationContainer container, final Map<Object, Object> context)
      throws ADLException {
    final Implementation impl = container.getImplementation();
    if (impl == null) return;

    final String className = impl.getClassName();
    if (className == null) {
      throw new ADLException("Implementation class name missing", impl);
    }
    Object c;
    try {
      c = implementationCodeLoaderItf.loadImplementation(className, null,
          context);
    } catch (final ImplementationException e) {
      throw new ADLException("Invalid implementation", impl, e);
    }
    impl.astSetDecoration("code", c);

    if (!(c instanceof Class)) return;

    final Class<?> implemClass = (Class<?>) c;
    if (container instanceof InterfaceContainer) {
      for (final Interface itf : ((InterfaceContainer) container)
          .getInterfaces()) {
        if (!(itf instanceof TypeInterface)) continue;

        final TypeInterface tItf = (TypeInterface) itf;
        if (tItf.getRole().equals(TypeInterface.SERVER_ROLE)) {
          final Object d = interfaceLoaderItf.loadInterface(
              tItf.getSignature(), context);
          if (d instanceof Class) {
            if (!((Class<?>) d).isAssignableFrom(implemClass)) {
              throw new ADLException(
                  "The implementation class does not implement "
                      + "all the server interfaces of the component", impl);
            }
          }
        }
      }
    }
    if (container instanceof AttributesContainer) {
      final Attributes attrs = ((AttributesContainer) container)
          .getAttributes();
      if (attrs != null) {
        final Object d = interfaceLoaderItf.loadInterface(attrs.getSignature(),
            context);
        if (d instanceof Class) {
          if (!((Class<?>) d).isAssignableFrom(implemClass)) {
            throw new ADLException(
                "The implementation class does not implement "
                    + "all the attribute controller interface of the component",
                impl);
          }
        }
      }
    }
  }

  protected void checkControllerContainer(final ControllerContainer container)
      throws ADLException {
    final Controller ctrl = container.getController();
    if (ctrl != null) {
      if (ctrl.getDescriptor() == null) {
        throw new ADLException("Controller descriptor missing", ctrl);
      }
    }
  }

  @Override
  public void bindFc(final String s, final Object o)
      throws NoSuchInterfaceException {

    if (s == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (LOADER_BINDING.equals(s)) {
      this.clientLoader = (Loader) o;
    } else if (INTERFACE_LOADER_BINDING.equals(s)) {
      this.interfaceLoaderItf = (IDLLoader) o;
    } else if (IMPLEMENTATION_LOADER_BINDING.equals(s)) {
      this.implementationCodeLoaderItf = (ImplementationCodeLoader) o;
    } else {
      throw new NoSuchInterfaceException("There is no interface named '" + s
          + "'");
    }
  }

  @Override
  public String[] listFc() {
    return new String[]{LOADER_BINDING, INTERFACE_LOADER_BINDING,
        IMPLEMENTATION_LOADER_BINDING};
  }

  @Override
  public Object lookupFc(final String s) throws NoSuchInterfaceException {

    if (s == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    } else if (LOADER_BINDING.equals(s)) {
      return this.clientLoader;
    } else if (INTERFACE_LOADER_BINDING.equals(s)) {
      return this.interfaceLoaderItf;
    } else if (IMPLEMENTATION_LOADER_BINDING.equals(s)) {
      return this.implementationCodeLoaderItf;
    } else {
      throw new NoSuchInterfaceException("There is no interface named '" + s
          + "'");
    }

  }

  @Override
  public void unbindFc(final String s) throws NoSuchInterfaceException {

    if (s == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (LOADER_BINDING.equals(s)) {
      this.clientLoader = null;
    } else if (INTERFACE_LOADER_BINDING.equals(s)) {
      this.interfaceLoaderItf = null;
    } else if (IMPLEMENTATION_LOADER_BINDING.equals(s)) {
      this.implementationCodeLoaderItf = null;
    } else {
      throw new NoSuchInterfaceException("There is no interface named '" + s
          + "'");
    }
  }
}
