/***
 * Fractal ADL Parser
 * Copyright (C) 2002-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 *
 * Contributor(s): Philippe Merle
 *                 Matthieu Leclercq (externalize interface loading)
 */

package org.objectweb.fractal.adl.types;

import java.util.Map;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.AbstractLoader;
import org.objectweb.fractal.adl.Definition;
import org.objectweb.fractal.adl.Loader;
import org.objectweb.fractal.adl.components.Component;
import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.adl.interfaces.IDLException;
import org.objectweb.fractal.adl.interfaces.IDLLoader;
import org.objectweb.fractal.adl.interfaces.Interface;
import org.objectweb.fractal.adl.interfaces.InterfaceContainer;

/**
 * A {@link org.objectweb.fractal.adl.Loader} to check {@link TypeInterface}
 * nodes in definitions. This loader checks that the Java interfaces specified
 * in these nodes exist.
 */

public class TypeLoader extends AbstractLoader {

  // --------------------------------------------------------------------------
  // Client interfaces
  // --------------------------------------------------------------------------
  public final static String INTERFACE_LOADER_BINDING = "interface-loader";

  public IDLLoader           interfaceCodeLoaderItf;

  // --------------------------------------------------------------------------
  // Implementation of the Loader interface
  // --------------------------------------------------------------------------

  public Definition load(final String name, final Map<Object, Object> context)
      throws ADLException {
    final Definition d = clientLoader.load(name, context);
    checkNode(d, context);
    return d;
  }

  // --------------------------------------------------------------------------
  // Checking methods
  // --------------------------------------------------------------------------

  private void checkNode(final Object node, final Map<Object, Object> context)
      throws ADLException {
    if (node instanceof InterfaceContainer) {
      checkInterfaceContainer((InterfaceContainer) node, context);
    }
    if (node instanceof ComponentContainer) {
      for (final Component comp : ((ComponentContainer) node).getComponents()) {
        checkNode(comp, context);
      }
    }
  }

  protected void checkInterfaceContainer(final InterfaceContainer container,
      final Map<Object, Object> context) throws ADLException {
    for (final Interface itf : container.getInterfaces()) {
      if (itf instanceof TypeInterface) {
        final String signature = ((TypeInterface) itf).getSignature();
        if (signature == null) {
          throw new ADLException("Signature missing", itf);
        } else {
          try {
            interfaceCodeLoaderItf.loadInterface(signature, context);
          } catch (final IDLException e) {
            throw new ADLException("Invalid signature '" + signature + "'",
                itf, e);
          }
        }
        final String role = ((TypeInterface) itf).getRole();
        if (role == null) {
          throw new ADLException("Role missing", itf);
        } else {
          if (!role.equals(TypeInterface.CLIENT_ROLE)
              && !role.equals(TypeInterface.SERVER_ROLE)) {
            throw new ADLException("Invalid role '" + role + "'", itf);
          }
        }
        final String contingency = ((TypeInterface) itf).getContingency();
        if (contingency != null) {
          if (!contingency.equals(TypeInterface.MANDATORY_CONTINGENCY)
              && !contingency.equals(TypeInterface.OPTIONAL_CONTINGENCY)) {
            throw new ADLException("Invalid contingency '" + contingency + "'",
                itf);
          }
        }
        final String cardinality = ((TypeInterface) itf).getCardinality();
        if (cardinality != null) {
          if (!cardinality.equals(TypeInterface.SINGLETON_CARDINALITY)
              && !cardinality.equals(TypeInterface.COLLECTION_CARDINALITY)) {
            throw new ADLException("Invalid cardinality '" + cardinality + "'",
                itf);
          }
        }
      }
    }
  }

  @Override
  public void bindFc(final String s, final Object o) {
    if (LOADER_BINDING.equals(s)) {
      this.clientLoader = (Loader) o;
    }
    if (INTERFACE_LOADER_BINDING.equals(s)) {
      this.interfaceCodeLoaderItf = (IDLLoader) o;
    }
  }

  @Override
  public String[] listFc() {
    return new String[]{LOADER_BINDING, INTERFACE_LOADER_BINDING};
  }

  @Override
  public Object lookupFc(final String s) {
    if (LOADER_BINDING.equals(s)) {
      return this.clientLoader;
    }
    if (INTERFACE_LOADER_BINDING.equals(s)) {
      return this.interfaceCodeLoaderItf;
    }
    return null;
  }

  @Override
  public void unbindFc(final String s) {
    if (LOADER_BINDING.equals(s)) {
      this.clientLoader = null;
    }
    if (INTERFACE_LOADER_BINDING.equals(s)) {
      this.interfaceCodeLoaderItf = null;
    }
  }

}
