/***
 * Fractal ADL Parser
 * Copyright (C) 2005 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Ali Erdem Ozcan
 */

package org.objectweb.fractal.adl.util;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.Map;
import java.util.logging.Formatter;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.LogRecord;
import java.util.logging.Logger;
import java.util.logging.StreamHandler;

/**
 * <p>
 * Utility class for logging messages. Using this class, one can create new
 * loggers or access existing loggers. Each logger prints its output to the file
 * system and to the console.
 * </p>
 * <p>
 * The verbose level for each logger can be set separately with a
 * {@link System#getProperty(String) system property} whose name is
 * <code>"&lt;name&gt;.&lt;device&gt;.level"</code> where
 * <code>&lt;name&gt;</code> stands for the name of the logger, and
 * <code>&lt;device&gt;</code> stands for either <code>"file"</code> or
 * <code>"console"</code>. Default verbose levels can be defined using
 * <code>"default.&lt;device&gt;.level"</code> properties. The loggers with no
 * level specified inherit from their parent. The highest level loggers inherit
 * from the default level. If no default level is specified, it is considered as
 * {@link Level#INFO}.
 * </p>
 * <p>
 * The output directory for loggers can be set with a
 * {@link System#getProperty(String) system property} called
 * <code>"loggers.outdir"</code>. If no output directory is specified, the
 * default value is considered as <code>"target/log"</code>.
 * </p>
 */
public final class FractalADLLogManager {
  private FractalADLLogManager() {
  }

  /** Default output directory. */
  public static final String                      DEFAULT_OUTPUT_DIR                  = "target/log";

  /** Default output file for the logger. */
  public static final String                      OUTPUT_FILE_NAME                    = "fractaladl.log.txt";

  /**
   * The name of the {@link System#getProperty(String) system property} used to
   * select the output directory for the loggers.
   */
  public static final String                      OUTPUT_DIR_PROPERTY_NAME            = "loggers.outdir";

  /**
   * The suffix of the {@link System#getProperty(String) system property} used
   * to set the properties of the console logger.
   */
  public static final String                      CONSOLE_LOGGER_SUFFIX               = ".console";
  /**
   * The suffix of the {@link System#getProperty(String) system property} used
   * to set the properties of the console logger.
   */
  public static final String                      FILE_LOGGER_SUFFIX                  = ".file";
  /**
   * The suffix of the {@link System#getProperty(String) system property} for
   * setting the verbose level of the loggers.
   */
  public static final String                      LEVEL_SUFFIX                        = ".level";

  /**
   * The name of the {@link System#getProperty(String) system property} for
   * setting the default verbose level for the console handler.
   */
  public static final String                      DEFAULT_CONSOLE_LEVEL_PROPERTY_NAME = "default"
                                                                                          + CONSOLE_LOGGER_SUFFIX
                                                                                          + LEVEL_SUFFIX;
  /**
   * The name of the {@link System#getProperty(String) system property} for
   * setting the default verbose level for the file handler.
   */
  public static final String                      DEFAULT_FILE_LEVEL_PROPERTY_NAME    = "default"
                                                                                          + FILE_LOGGER_SUFFIX
                                                                                          + LEVEL_SUFFIX;
  /**
   * The name of logger which is the parent of all loggers created by this log
   * manager.
   */
  public static final String                      ROOT_LOGGER_NAME                    = "fractaladl";
  /**
   * The default verbose level.
   */
  public static final Level                       DEFAULT_VERBOSE_LEVEL               = Level.INFO;

  /** A cache containing already created loggers. */
  protected static final Map<String, Logger>      LOGGER_MAP                          = new HashMap<String, Logger>();

  /**
   * A default logger for logging the step-by-step continuation of the
   * compilation.
   */
  public static final Logger                      STEP;

  protected static final Map<Logger, LoggerShelf> LOGGER_SHELF_MAP;

  static {
    // Get the default console and file verbose levels
    final String consoleLevelString = System.getProperty(
        DEFAULT_CONSOLE_LEVEL_PROPERTY_NAME, DEFAULT_VERBOSE_LEVEL.toString());
    Level consoleLevel;
    try {
      consoleLevel = Level.parse(consoleLevelString);
    } catch (final IllegalArgumentException e) {
      System.err.println("Warning: Invalid level value for ["
          + DEFAULT_CONSOLE_LEVEL_PROPERTY_NAME + "]: \"" + consoleLevelString
          + "\". Defaulting to the parent's verbose level");
      consoleLevel = DEFAULT_VERBOSE_LEVEL;
    }
    // Find the file verbose level of the logger
    final String fileLevelString = System.getProperty(
        DEFAULT_FILE_LEVEL_PROPERTY_NAME, DEFAULT_VERBOSE_LEVEL.toString());
    Level fileLevel;
    try {
      fileLevel = Level.parse(fileLevelString);
    } catch (final IllegalArgumentException e) {
      System.err.println("Warning: Invalid level value for ["
          + DEFAULT_FILE_LEVEL_PROPERTY_NAME + "]: \"" + consoleLevelString
          + "\". Defaulting to the parent's verbose level");
      fileLevel = DEFAULT_VERBOSE_LEVEL;
    }

    LOGGER_SHELF_MAP = new IdentityHashMap<Logger, LoggerShelf>();
    LOGGER_MAP.put(ROOT_LOGGER_NAME, new LoggerShelf(ROOT_LOGGER_NAME,
        consoleLevel, fileLevel).getLogger());
    STEP = getLogger("step");
  }

  /**
   * Find or create a logger for a named subsystem that prints its output to the
   * file system and to the console. The verbose level for those output devices
   * can be separately set with a
   * {@link System#getProperty(String) system property} whose name is
   * <code>"&lt;name&gt;.&lt;device&gt;.level"</code> where
   * <code>&lt;name&gt;</code> stands for the name of the logger, and
   * <code>&lt;device&gt;</code> stands for either <code>"file"</code> or
   * <code>"console"</code>. The loggers with no level specified inherit from
   * their parent. The highest level loggers inherit from the default level.
   * 
   * @param name the name of the logger
   * @return a suitable {@link Logger}.
   */
  public static Logger getLogger(final String name) {
    // Look into the logger cache
    if (LOGGER_MAP.containsKey(name)) {
      // If a logger for that name is found, return it.
      return LOGGER_MAP.get(name);
    }
    // Otherwise, create a new logger and register it in the cache.
    final Logger logger = new LoggerShelf(name).getLogger();
    LOGGER_MAP.put(name, logger);
    return logger;
  }

  /**
   * Find or create a logger for a named subsystem that prints its output to the
   * file system and to the console with the specification of the verbose level
   * for each device.
   * 
   * @param name the name of the logger
   * @param consoleLevel the verbose level of the console device.
   * @param fileLevel the verbose elvel of the file device.
   * @return a suitable {@link Logger}.
   */
  public static Logger getLogger(final String name, final Level consoleLevel,
      final Level fileLevel) {
    // Look into the logger cache
    if (LOGGER_MAP.containsKey(name)) {
      // If a logger for that name is found, return it.
      return LOGGER_MAP.get(name);
    }
    // Otherwise, create a new logger and register it in the cache.
    final Logger logger = new LoggerShelf(name, consoleLevel, fileLevel)
        .getLogger();
    LOGGER_MAP.put(name, logger);
    return logger;
  }

  /**
   * A {@link Formatter} for printing very simple text messages on the console.
   */
  private static class SimpleConsoleFormatter extends Formatter {

    @Override
    public String format(final LogRecord record) {
      final String loggerName = record.getLoggerName();
      return "[" + loggerName.substring(loggerName.indexOf('.') + 1) + "] "
          + record.getMessage() + "\n";
    }
  }

  /**
   * A {@link StreamHandler} that prints on {@link System#out}.
   */
  public static class StdOutConsoleHandler extends StreamHandler {
    StdOutConsoleHandler() {
      super(System.out, new SimpleConsoleFormatter());
    }

    @Override
    public synchronized void publish(final LogRecord record) {
      super.publish(record);
      flush();
    }
  }

  /**
   * A {@link StreamHandler} that prints on the file system..
   */
  public static class SingletonFileHandler extends StreamHandler {
    private static PrintStream outputStream;
    protected static String    outputDirectory = null;
    static {
      // Create the output directory for the loggers
      if (outputDirectory == null) {
        outputDirectory = System.getProperty(OUTPUT_DIR_PROPERTY_NAME,
            DEFAULT_OUTPUT_DIR);
        new File(outputDirectory).mkdirs();
      }
      // Create the output stream
      try {
        outputStream = new PrintStream(new File(outputDirectory,
            OUTPUT_FILE_NAME));
      } catch (final FileNotFoundException e) {
        System.err.println("Log file '" + outputDirectory + File.pathSeparator
            + OUTPUT_FILE_NAME + "' not found");
        e.printStackTrace();
      }
    }

    public SingletonFileHandler() {
      super(outputStream, new SimpleConsoleFormatter());
    }

    @Override
    public synchronized void publish(final LogRecord record) {
      super.publish(record);
      flush();
    }
  }

  private static class LoggerShelf {
    private final Logger  logger;

    private final Logger  parentLogger;

    private final Handler consoleHandler;

    private final Handler fileHandler;

    public LoggerShelf(final String name) {
      final String loggerName = name.equals(ROOT_LOGGER_NAME)
          ? name
          : ROOT_LOGGER_NAME + "." + name;
      // Create the logger and register it in the map
      logger = Logger.getLogger(loggerName);
      LOGGER_SHELF_MAP.put(logger, this);
      // Find its parent
      parentLogger = logger.getParent();
      final LoggerShelf parentLoggerShelf = LOGGER_SHELF_MAP.get(parentLogger);
      // Find the console verbose level of the logger
      final String consoleLevelString = System.getProperty(name
          + CONSOLE_LOGGER_SUFFIX + LEVEL_SUFFIX, parentLoggerShelf
          .getConsoleHandler().getLevel().toString());
      Level consoleLevel;
      try {
        consoleLevel = Level.parse(consoleLevelString);
      } catch (final IllegalArgumentException e) {
        System.err.println("Warning: Invalid level value for [" + name
            + CONSOLE_LOGGER_SUFFIX + LEVEL_SUFFIX + "]: \""
            + consoleLevelString
            + "\". Defaulting to the parent's verbose level");
        consoleLevel = parentLoggerShelf.getConsoleHandler().getLevel();
      }
      // Find the file verbose level of the logger
      final String fileLevelString = System.getProperty(name
          + FILE_LOGGER_SUFFIX + LEVEL_SUFFIX, parentLoggerShelf
          .getFileHandler().getLevel().toString());
      Level fileLevel;
      try {
        fileLevel = Level.parse(fileLevelString);
      } catch (final IllegalArgumentException e) {
        System.err.println("Warning: Invalid level value for [" + name
            + FILE_LOGGER_SUFFIX + LEVEL_SUFFIX + "]: \"" + consoleLevelString
            + "\". Defaulting to the parent's verbose level");
        fileLevel = parentLoggerShelf.getFileHandler().getLevel();
      }
      // Default configuration of the logger
      logger.setUseParentHandlers(false);
      // Set the console handler
      consoleHandler = new StdOutConsoleHandler();
      consoleHandler.setLevel(consoleLevel);
      logger.addHandler(consoleHandler);
      // Set the file handler
      fileHandler = new SingletonFileHandler();
      fileHandler.setLevel(fileLevel);
      logger.addHandler(fileHandler);

      // Set the logger's verbose level to the most verbose one among the
      // console and the file
      logger.setLevel(consoleLevel.intValue() < fileLevel.intValue()
          ? consoleLevel
          : fileLevel);
    }

    public LoggerShelf(final String name, final Level consoleLevel,
        final Level fileLevel) {
      final String loggerName = name.equals(ROOT_LOGGER_NAME)
          ? name
          : ROOT_LOGGER_NAME + "." + name;
      // Create the logger and register it in the map
      logger = Logger.getLogger(loggerName);
      LOGGER_SHELF_MAP.put(logger, this);
      // Find its parent
      parentLogger = logger.getParent();
      // Default configuration of the logger
      logger.setUseParentHandlers(false);
      // Set the console handler
      consoleHandler = new StdOutConsoleHandler();
      consoleHandler.setLevel(consoleLevel);
      logger.addHandler(consoleHandler);
      // Set the file handler
      fileHandler = new SingletonFileHandler();
      fileHandler.setLevel(fileLevel);
      logger.addHandler(fileHandler);

      // Set the logger's verbose level to the most verbose one among the
      // console and the file
      logger.setLevel(consoleLevel.intValue() < fileLevel.intValue()
          ? consoleLevel
          : fileLevel);
    }

    public Logger getLogger() {
      return logger;
    }

    public Handler getConsoleHandler() {
      return consoleHandler;
    }

    public Handler getFileHandler() {
      return fileHandler;
    }
  }

}
